import * as vscode from 'vscode';
import axios from 'axios';

let config: {
    apiEndpoint: string;
    apiAuthentication: string;
    apiModel: string;
    apiTemperature: number;
    numPredict: number;
    promptWindowSize: number;
    completionKeys: string[];
    responsePreview: boolean;
    responsePreviewMaxTokens: number;
    responsePreviewDelay: number;
    continueInline: boolean;
    keepAlive: number;
    topP: number;
};

let previewDecorationType: vscode.TextEditorDecorationType;

function updateConfig() {
    const vsConfig = vscode.workspace.getConfiguration('fabelous-autocoder');
    config = {
        apiEndpoint: vsConfig.get('endpoint') || 'http://localhost:11434/api/generate',
        apiAuthentication: vsConfig.get('authentication') || '',
        apiModel: vsConfig.get('model') || 'fabelous-coder:latest',
        apiTemperature: vsConfig.get('temperature') || 0.7,
        numPredict: vsConfig.get('max tokens predicted') || 1000,
        promptWindowSize: vsConfig.get('prompt window size') || 2000,
        completionKeys: (vsConfig.get('completion keys') as string || ' ').split(''),
        responsePreview: vsConfig.get('response preview') || false,
        responsePreviewMaxTokens: vsConfig.get('preview max tokens') || 50,
        responsePreviewDelay: vsConfig.get('preview delay') || 0,
        continueInline: vsConfig.get('continue inline') || false,
        keepAlive: vsConfig.get('keep alive') || 30,
        topP: vsConfig.get('top p') || 1,
    };
}

function createPreviewDecorationType() {
    previewDecorationType = vscode.window.createTextEditorDecorationType({
        after: {
            color: '#888888',
            fontStyle: 'italic',
        },
        textDecoration: 'none; display: none;',
    });
}

function getContextLines(document: vscode.TextDocument, position: vscode.Position): string {
    const startLine = Math.max(0, position.line - 1);
    const endLine = position.line;
    return document.getText(new vscode.Range(startLine, 0, endLine, position.character));
}

function createFIMPrompt(prefix: string, language: string): string {
    return `<fim_prefix>${prefix}<fim_middle><fim_suffix>${language}\n`;
}

async function generateCompletion(prompt: string, cancellationToken: vscode.CancellationToken): Promise<string> {
    const axiosCancelToken = new axios.CancelToken((c) => {
        cancellationToken.onCancellationRequested(() => c('Request cancelled'));
    });

    const response = await axios.post(config.apiEndpoint, {
        model: config.apiModel,
        prompt: prompt,
        stream: false,
        raw: true,
        options: {
            num_predict: config.numPredict + 100, // Generate extra lines for preview
            temperature: config.apiTemperature,
            stop: ['<fim_suffix>', '```'],
            keep_alive: config.keepAlive,
            top_p: config.topP,
        }
    }, {
        cancelToken: axiosCancelToken,
        headers: {
            'Authorization': config.apiAuthentication
        }
    });

    const fullCompletion = response.data.response.replace(/<fim_middle>|<fim_suffix>|<fim_prefix>/g, '').trim();
    const completionLines = fullCompletion.split('\n');
    const usedContextLines = prompt.split('\n').length;
    
    // Remove used context and take extra lines for preview
    return completionLines.slice(usedContextLines, usedContextLines + config.numPredict).join('\n');
}

class CompletionManager {
    private textEditor: vscode.TextEditor;
    private document: vscode.TextDocument;
    private startPosition: vscode.Position;
    private completionText: string;
    private previewLines: string[];

    constructor(textEditor: vscode.TextEditor, startPosition: vscode.Position, completionText: string) {
        this.textEditor = textEditor;
        this.document = textEditor.document;
        this.startPosition = startPosition;
        this.completionText = completionText;
        this.previewLines = completionText.split('\n');
    }

    public async showPreview() {
        const previewRanges: vscode.DecorationOptions[] = this.previewLines.map((line, index) => ({
            range: new vscode.Range(this.startPosition.translate(index, 0), this.startPosition.translate(index, 0)),
            renderOptions: {
                after: {
                    contentText: line,
                }
            }
        }));
        this.textEditor.setDecorations(previewDecorationType, previewRanges);
    }

    public async acceptCompletion() {
        const edit = new vscode.WorkspaceEdit();
        edit.insert(this.document.uri, this.startPosition, this.completionText);
        await vscode.workspace.applyEdit(edit);
        this.clearPreview();
    }

    public clearPreview() {
        this.textEditor.setDecorations(previewDecorationType, []);
    }
}

async function autocompleteCommand(textEditor: vscode.TextEditor, edit: vscode.TextEditorEdit, ...args: any[]) {
    const cancellationTokenSource = new vscode.CancellationTokenSource();
    const cancellationToken = cancellationTokenSource.token;

    try {
        const document = textEditor.document;
        const position = textEditor.selection.active;
        const context = getContextLines(document, position);
        const fimPrompt = createFIMPrompt(context, document.languageId);

        const completionText = await vscode.window.withProgress({
            location: vscode.ProgressLocation.Notification,
            title: 'Fabelous Autocoder',
            cancellable: true,
        }, async (progress, progressCancellationToken) => {
            progress.report({ message: 'Generating...' });
            return await generateCompletion(fimPrompt, progressCancellationToken);
        });

        const completionManager = new CompletionManager(textEditor, position, completionText);
        await completionManager.showPreview();

        const disposable = vscode.window.onDidChangeTextEditorSelection(async (event) => {
            if (event.textEditor !== textEditor) return;

            if (event.kind === vscode.TextEditorSelectionChangeKind.Keyboard) {
                await completionManager.acceptCompletion();
                disposable.dispose();
            }
        });
    } catch (err: any) {
        vscode.window.showErrorMessage(`Fabelous Autocoder encountered an error: ${err.message}`);
        console.error(err);
    } finally {
        cancellationTokenSource.dispose();
    }
}

async function provideCompletionItems(document: vscode.TextDocument, position: vscode.Position, cancellationToken: vscode.CancellationToken) {
    const item = new vscode.CompletionItem('Fabelous autocompletion');
    item.insertText = new vscode.SnippetString('${1:}');
    item.documentation = new vscode.MarkdownString('Press `Enter` to get an autocompletion from Fabelous Autocoder');

    if (config.responsePreview) {
        await new Promise(resolve => setTimeout(resolve, config.responsePreviewDelay * 1000));
        if (cancellationToken.isCancellationRequested) {
            return [item];
        }

        const context = getContextLines(document, position);
        const fimPrompt = createFIMPrompt(context, document.languageId);

        try {
            const previewText = await generateCompletion(fimPrompt, cancellationToken);
            item.detail = previewText.split('\n')[0]; // Show first line as preview
        } catch (error) {
            console.error('Error fetching preview:', error);
        }
    }

    if (config.continueInline || !config.responsePreview) {
        item.command = {
            command: 'fabelous-autocoder.autocomplete',
            title: 'Fabelous Autocomplete',
            arguments: []
        };
    }

    return [item];
}

export function activate(context: vscode.ExtensionContext) {
    updateConfig();
    createPreviewDecorationType();

    context.subscriptions.push(
        vscode.workspace.onDidChangeConfiguration(updateConfig),
        vscode.languages.registerCompletionItemProvider('*', { provideCompletionItems }, ...config.completionKeys),
        vscode.commands.registerTextEditorCommand('fabelous-autocoder.autocomplete', autocompleteCommand)
    );
}

export function deactivate() {}
